%
% Performs selection of GreedyTL model using efficient leave-one-out objective.
%
% select_greedytl_loo(hp, x, source_pred, y);
%
% Input:
% hp.lambda -- Tikhonov regularization parameter.
% x -- m-by-d matrix of training examples, where m is number of examples and d is the dimension.
% source_pred -- n-by-d matrix of source hypothesis predictions on training examples, where n is number of source hypotheses.
% y -- vector of training labels of size m.
%
% Output:
% min_lambda -- best regularization parameter found.
%
% Contact the author: ilja.kuzborskij [at] idiap.ch
%                     http:%idiap.ch/~ikuzbor
%
% ----------------------------------------------------------------------------------
%
% Copyright (c) 2014 Idiap Research Institute, http:%www.idiap.ch/
% Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>
%
% This file is part of GreedyTL.
%
% GreedyTL is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License version 3 as
% published by the Free Software Foundation.
%
% GreedyTL is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with GreedyTL. If not, see <http:%www.gnu.org/licenses/>.
%

function min_lambda = select_greedytl_loo(hp, x, source_pred, y)
   joint_train_x = [source_pred x];
   joint_train_x = standardize_cols(joint_train_x);
   y = standardize_cols(y);

   m = numel(y);

   min_err = 0;
   min_lambda = 0;

    for j = 1:numel(hp.lambda)
        fprintf('select_greedytl_loo:: Model selection %d/%d...\n', j, numel(hp.lambda));
        l = hp.lambda(j);
        hp_ = hp;
        hp_.lambda = l;

        weights = train_fr_tikhonov_mex(hp_, joint_train_x, y);
        S = find(weights~=0);

        Kr = joint_train_x(:, S)*joint_train_x(:, S)' + l*eye(m);
        invKr = inv(Kr);
        a = invKr * y;

        err = mean( (diag(1./diag(invKr)) * a).^2 )

        if err < min_err
            min_err = err;
            min_lambda = l;
        end
    end;
end
