//
// This function performs Tikhonov (L2) regularized forward regression.
//
// train_fr_tikhonov(mat X, vec y, double errtol, double lambda);
//
// Input:
// X -- m-by-d (Eigen) matrix of training examples, where m is number of examples and d is the dimension.
// y -- vector of training labels of size m.
// errtol -- greedy selection stopping threshold.
// lambda -- Tikhonov regularization parameter.

//
// Output:
// w -- weight vector of dimension d.
//
// Contact the author: ilja.kuzborskij [at] idiap.ch
//                     http://idiap.ch/~ikuzbor
//
// ----------------------------------------------------------------------------------
//
// Copyright (c) 2014 Idiap Research Institute, http://www.idiap.ch/
// Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>
//
// This file is part of GreedyTL.
//
// GreedyTL is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3 as
// published by the Free Software Foundation.
//
// GreedyTL is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GreedyTL. If not, see <http://www.gnu.org/licenses/>.
//

#include "train_fr_tikhonov.hpp"
#include <iostream>

using namespace std;

vec train_fr_tikhonov(mat X, vec y, double errtol, double lambda) {
    size_t m = X.rows();
    size_t d = X.cols();
    size_t active_d = d;
    vec U = vec::Zero(d);

    mat G = (1/(m*lambda)) * mat::Identity(m, m);
    mat xixiT = mat::Zero(m,m);
    mat xSxS = mat::Zero(m,m);
    vec xi;
    mat G_;

    double prev_acc = std::numeric_limits<double>::max();
    double max_acc = 0;
    size_t max_i = -1;

    while (active_d > 0) {
        max_acc = 0;
        max_i = -1;

        for (size_t i = 0; i < d; ++i) {
            if (0 == U[i]) {
                xi = X.col(i);
                xixiT = xi*xi.transpose();

                double acc = (1.0/m) * y.transpose() *
                    (G - (G.transpose() * xixiT * G) / (1 + xi.transpose() * G * xi))
                    * (xSxS + xixiT) * y;

                if (acc > max_acc) {
                    max_acc = acc;
                    max_i = i;
                }
            }
        }

        if (max_acc == 0) {
            std::cerr << "train_fr_tikhonov_eigen:: Achieved accuracy 0." << std::endl;
            break;
        }

        // Found the best member indexed by max_i
        U[max_i] = 1;
        active_d--;
        xi = X.col(max_i);
        xixiT = xi*xi.transpose();
        // Updating the inverse
        G = G - (G.transpose() * xixiT * G) / (1 + xi.transpose() * G * xi);

        // Accounting for another member
        xSxS += xixiT;

        if (fabs(prev_acc - max_acc) < errtol)
            break;
        else
            prev_acc = max_acc;

    }

    // Solving here
    return U.asDiagonal() * X.transpose() * G * y;
}
