#include <mex.h>
#include <math.h>
#include <string.h>

#include "train_fr_tikhonov.hpp"

#include <iostream>

// Debugging includes
#include <fstream>
//

#define GETELEMENT(mat, row, col, colsize)   mat[col*colsize+row]
#define GETPTR(mat, idx, size)           (mat+idx*size)

char* expect_string(const mxArray* structure, char* key) {
    mxArray *field;
    field = mxGetField(structure, 0, key);
    if(field == NULL || !mxIsChar(field) || mxGetM(field) != 1) {
        mexErrMsgIdAndTxt("MATLAB:train_fr_tikhonov_mex",
                          "Expected to read a string from the struct.");
    }
    return mxArrayToString(field);
}

double expect_double(const mxArray* structure, char* key) {
    mxArray *field;
    field = mxGetField(structure, 0, key);
    if(field == NULL || !mxIsNumeric(field) || mxGetM(field) != 1 || mxGetN(field) != 1) {
        mexErrMsgIdAndTxt("MATLAB:train_fr_tikhonov_mex:expectDouble",
                          "Expecting to read a double from the struct.");
    }
    return *((double*) mxGetData(field));
}

mat double_to_eigen_mat(double *X, mwSignedIndex m, mwSignedIndex n) {
    mat M(m, n);
    for (mwSignedIndex i=0; i < m; ++i) {
        for (mwSignedIndex j=0; j < n; ++j) {
            M(i,j) = GETELEMENT(X, i, j, m);
        }
    }
    return M;
}

vec double_to_eigen_vec(double *y, mwSignedIndex m) {
    vec v(m);

    for (mwSignedIndex i=0; i < m; ++i) {
        v(i) = y[i];
    }
    return v;
}

void eigen_vec_to_double(double *out, vec &v) {
    for (mwSignedIndex i=0; i < v.size(); ++i) {
        out[i] = v(i);
    }
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    mwSignedIndex m, d;
    double* X;
    double* y;
    double* beta;
    const mxArray* config;
    double errtol, lambda;

    /* Check I/O number */
    if(nlhs!=1) {
	mexErrMsgIdAndTxt("MATLAB:ckernel:lhs",
			  "This function provides 1 output matrix.");
    }
    if(nrhs!=3) {
	mexErrMsgIdAndTxt("MATLAB:ckernel:rhs",
			  "This function requires configuration struct, ones matrix and one vector.");
    }

    if(mxGetM(prhs[1]) != mxGetM(prhs[2])) {
	mexErrMsgIdAndTxt("MATLAB:ckernel:rhs",
			  "This function requires X and y to contain an equal number of rows.");
    }

    if(!mxIsStruct(prhs[0])) {
	mexErrMsgIdAndTxt("MATLAB:ckernel:rhs",
			  "This function requires the first parameter to be a configuration struct.");
    }

    m = (mwSignedIndex) mxGetM(prhs[1]);
    d = (mwSignedIndex) mxGetN(prhs[1]);

    X = mxGetPr(prhs[1]);
    y = mxGetPr(prhs[2]);

    config = prhs[0];

    errtol = expect_double(config, "errtol");
    lambda = expect_double(config, "lambda");

    mat matX = double_to_eigen_mat(X, m, d);
    vec vecy = double_to_eigen_vec(y, m);

    vec vecbeta = train_fr_tikhonov(matX, vecy, errtol, lambda);

    plhs[0] = mxCreateDoubleMatrix(d, 1, mxREAL);
    beta = mxGetPr(plhs[0]);
    memset((void*) beta, 0, d*sizeof(mxREAL));

    eigen_vec_to_double(beta, vecbeta);

    return;
}
